@@ -4,8 +4,12 @@ from torch import nn
 import math
 from operator import mul
 from fractions import gcd
+from collections import namedtuple
 from functools import partial, reduce
 
+from local_attention import LocalAttention
+from linformer import LinformerSelfAttention
+
 from product_key_memory import PKM
 from axial_positional_embedding import AxialPositionalEmbedding
 from linear_attention_transformer.reversible import ReversibleSequence, SequentialSequence
@@ -14,6 +18,12 @@ from linear_attention_transformer.reversible import ReversibleSequence, Sequenti
 
 DEFAULT_PSI = lambda x: F.elu(x) + 1
 
+
+# namedtuple settings
+
+LinformerSettings = namedtuple('LinformerSettings', ['k'])
+LinformerContextSettings = namedtuple('LinformerContextSettings', ['seq_len', 'k'])
+
 # helper functions
 
 def default(value, d):
@@ -34,8 +44,9 @@ def merge_dims(ind_from, ind_to, tensor):
     shape[arr_slice] = [reduce(mul, shape[arr_slice])]
     return tensor.reshape(*shape)
 
-def expand_dim(t, dim, k):
-    t = t.unsqueeze(dim)
+def expand_dim(t, dim, k, unsqueeze=True):
+    if unsqueeze:
+        t = t.unsqueeze(dim)
     expand_shape = [-1] * len(t.shape)
     expand_shape[dim] = k
     return t.expand(*expand_shape)
@@ -97,83 +108,6 @@ class AbsolutePositionalEmbedding(nn.Module):
         t = torch.arange(x.shape[1], device=x.device)
         return self.emb(t)
 
-# local attention
-
-def look_around(x, backward = 1, forward = 0, pad_value = -1, dim = 2):
-    t = x.shape[1]
-    dims = (len(x.shape) - dim) * (0, 0)
-    padded_x = F.pad(x, (*dims, backward, forward), value= pad_value)
-    tensors = [padded_x[:, ind:(ind + t), ...] for ind in range(forward + backward + 1)]
-    return torch.cat(tensors, dim=dim)
-
-class LocalAttention(nn.Module):
-    def __init__(self, bucket_size, heads, head_dim, causal = False, look_backward = 1, look_forward = None, dropout = 0.):
-        super().__init__()
-        self.look_forward = default(look_forward, 0 if causal else 1)
-        assert not (causal and self.look_forward > 0), 'you cannot look forward if causal'
-
-        self.bucket_size = bucket_size
-        self.causal = causal
-        self.look_backward = look_backward
-
-        self.heads = heads
-        self.dropout = nn.Dropout(dropout)
-
-    def forward(self, q, k, v, input_mask = None):
-        shape = q.shape
-
-        merge_into_batch = lambda t: t.reshape(-1, *t.shape[-2:])
-        q, k, v = map(merge_into_batch, (q, k, v))
-
-        b, t, e, h, device, dtype = *q.shape, self.heads, q.device, q.dtype
-        bucket_size, causal, look_backward, look_forward = self.bucket_size, self.causal, self.look_backward, self.look_forward
-        assert (t % bucket_size) == 0, f'sequence length {t} must be divisible by bucket size {bucket_size} for local attention'
-
-        buckets = t // bucket_size
-
-        ticker = torch.arange(t, device=device, dtype=dtype)[None, :]
-        b_t = ticker.reshape(1, buckets, bucket_size)
-
-        bucket_fn = lambda t: t.reshape(b, buckets, bucket_size, -1)
-        bq, bk, bv = map(bucket_fn, (q, k, v))
-
-        look_around_kwargs = {'backward': look_backward, 'forward': look_forward}
-        bk = look_around(bk, **look_around_kwargs)
-        bv = look_around(bv, **look_around_kwargs)
-
-        bq_t = b_t
-        bq_k = look_around(b_t, **look_around_kwargs)
-
-        dots = torch.einsum('bhie,bhje->bhij', bq, bk) * (e ** -0.5)
-
-        mask_value = max_neg_value(dots)
-
-        if causal:
-            mask = bq_t[:, :, :, None] < bq_k[:, :, None, :]
-            dots.masked_fill_(mask, mask_value)
-            del mask
-
-        mask = bq_k[:, :, None, :] == -1
-        dots.masked_fill_(mask, mask_value)
-        del mask
-
-        if input_mask is not None:
-            h = b // input_mask.shape[0]
-            input_mask = input_mask.reshape(-1, buckets, bucket_size)
-            mq = mk = input_mask
-            mk = look_around(mk, pad_value=False, **look_around_kwargs)
-            mask = (mq[:, :, :, None] * mk[:, :, None, :])
-            mask = merge_dims(0, 1, expand_dim(mask, 1, h))
-            dots.masked_fill_(~mask, mask_value)
-            del mask
-
-        attn = dots.softmax(dim=-1)
-        attn = self.dropout(attn)
-
-        out = torch.einsum('bhij,bhje->bhie', attn, bv)
-        out = out.reshape(*shape)
-        return out
-
 # feedforward
 
 class GELU_(nn.Module):
@@ -208,10 +142,15 @@ class FeedForward(nn.Module):
 # self attention layer
 
 def linear_attn(q, k, v, kv_mask = None, one_kv_head = False):
+    dim = q.shape[-1]
+    (q, k) = map(lambda x: x * (dim ** -0.25), (q, k))
+
     if kv_mask is not None:
         mask_value = max_neg_value(q)
         mask = kv_mask[:, :, None] if one_kv_head else kv_mask[:, None, :, None]
         k = k.masked_fill_(~mask, mask_value)
+        v = v.masked_fill_(~mask, 0.)
+        del mask
 
     q = q.softmax(dim=-1)
     k = k.softmax(dim=-2)
@@ -229,12 +168,16 @@ def causal_linear_attn(q, k, v, kv_mask = None, psi = DEFAULT_PSI, one_kv_head =
     bucket_size = default(bucket_size, 64)
     assert (n % bucket_size) == 0, f'sequence length {n} must be divisible by the bucket size {bucket_size} for causal linear attention'
 
+    (q, k) = map(lambda x: x * (e ** -0.25), (q, k))
+
     q = q.softmax(dim=-1)
     k = psi(k)
 
     if kv_mask is not None:
         mask = kv_mask[:, :, None] if one_kv_head else kv_mask[:, None, :, None]
         k = k.masked_fill_(~mask, 0.)
+        v = v.masked_fill_(~mask, 0.)
+        del mask
 
     bucket_fn = lambda x: x.reshape(*x.shape[:-2], -1, bucket_size, e)
     b_q, b_k, b_v = map(bucket_fn, (q, k, v))
@@ -258,7 +201,7 @@ def causal_linear_attn(q, k, v, kv_mask = None, psi = DEFAULT_PSI, one_kv_head =
     return attn.reshape(*q.shape)
 
 class SelfAttention(nn.Module):
-    def __init__(self, dim, heads, causal = False, one_kv_head = False, psi_fn = DEFAULT_PSI, blindspot_size = 1, n_local_attn_heads = 0, local_attn_window_size = 128, receives_context = False):
+    def __init__(self, dim, heads, causal = False, one_kv_head = False, psi_fn = DEFAULT_PSI, blindspot_size = 1, n_local_attn_heads = 0, local_attn_window_size = 128, receives_context = False, dropout = 0., attn_dropout = 0.):
         super().__init__()
         assert (dim % heads) == 0, 'embedding dimension must be divisible by number of heads'
         d_heads = dim // heads
@@ -268,21 +211,23 @@ class SelfAttention(nn.Module):
         self.psi_fn = psi_fn
         self.receives_context = receives_context
 
+        self.global_attn_heads = heads - n_local_attn_heads
         self.global_attn_fn = linear_attn if not causal else partial(causal_linear_attn, psi=psi_fn, bucket_size = blindspot_size)
 
         self.local_attn_heads = n_local_attn_heads
-        self.local_attn  = LocalAttention(local_attn_window_size, n_local_attn_heads, d_heads, causal = causal)
+        self.local_attn  = LocalAttention(local_attn_window_size, causal = causal, dropout = attn_dropout)
 
         self.to_q = nn.Linear(dim, dim, bias = False)
 
-        kv_heads = 1 if one_kv_head else heads
+        kv_heads = (int(self.local_attn_heads > 0) + int(self.global_attn_heads > 0)) if one_kv_head else heads
+
         self.one_kv_head = one_kv_head
         self.kv_heads = kv_heads
         self.to_k = nn.Linear(dim, d_heads * kv_heads, bias = False)
         self.to_v = nn.Linear(dim, d_heads * kv_heads, bias = False)
 
         self.to_out = nn.Linear(dim, dim)
-
+        self.dropout = nn.Dropout(dropout)
 
     def forward(self, x, input_mask = None, context = None, context_mask = None, **kwargs):
         assert not (self.receives_context and context is None), 'context must be supplied if self attention is in receives context mode'
@@ -293,12 +238,10 @@ class SelfAttention(nn.Module):
             q, k, v = (self.to_q(x), self.to_k(context), self.to_v(context))
 
         b, t, e, h, dh = *q.shape, self.heads, self.d_heads
-        merge_heads = lambda x: x.reshape(b, -1, h, dh).transpose(1, 2)
 
-        q = merge_heads(q)
+        merge_heads = lambda x: x.reshape(*x.shape[:2], -1, dh).transpose(1, 2)
 
-        if not self.one_kv_head:
-            k, v = map(merge_heads, (k, v))
+        q, k, v = map(merge_heads, (q, k, v))
 
         out = []
 
@@ -308,8 +251,14 @@ class SelfAttention(nn.Module):
             (lq, q), (lk, k), (lv, v) = map(split_index_fn, (q, k, v))
         else:
             lq, q = split_index_fn(q)
-            lk = expand_dim(k, 1, self.local_attn_heads)
-            lv = expand_dim(v, 1, self.local_attn_heads)
+
+            split_kv_fn = partial(split_at_index, 1, int(self.local_attn_heads > 0))
+            (lk, k), (lv, v) = map(split_kv_fn, (k, v))
+
+            local_expand_heads_fn = lambda t: expand_dim(t, 1, self.local_attn_heads, unsqueeze=False)
+            lk, lv = map(local_expand_heads_fn, (lk, lv))
+
+            k, v = map(lambda t: t.squeeze(1), (k, v))
 
         has_local, has_global = map(lambda x: x.shape[1] > 0, (lq, q))
 
@@ -324,7 +273,7 @@ class SelfAttention(nn.Module):
 
         attn = torch.cat(out, dim=1)
         attn = attn.transpose(1, 2).reshape(b, t, -1)
-        return self.to_out(attn)
+        return self.dropout(self.to_out(attn))
 
 # transformer and language model classes
 
@@ -346,8 +295,12 @@ class FoldAxially(nn.Module):
         return x
 
 class LinearAttentionTransformer(nn.Module):
-    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, causal = False, one_kv_head = False, ff_chunks = 1, reversible = False, blindspot_size = 1, n_local_attn_heads = 0, local_attn_window_size = 128, psi_fn = DEFAULT_PSI, receives_context = False, attend_axially = False, pkm_layers = tuple(), pkm_num_keys = 128):
+    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, causal = False, one_kv_head = False, ff_chunks = 1, ff_glu = False, ff_dropout = 0., attn_layer_dropout = 0., attn_dropout = 0., reversible = False, blindspot_size = 1, n_local_attn_heads = 0, local_attn_window_size = 128, psi_fn = DEFAULT_PSI, receives_context = False, attend_axially = False, pkm_layers = tuple(), pkm_num_keys = 128, linformer_settings = None, context_linformer_settings = None):
         super().__init__()
+        assert not (causal and linformer_settings is not None), 'Linformer self attention layer can only be used for non-causal networks'
+        assert linformer_settings is None or isinstance(linformer_settings, LinformerSettings), 'Linformer self-attention settings must be a LinformerSettings namedtuple'
+        assert context_linformer_settings is None or isinstance(context_linformer_settings, LinformerContextSettings), 'Linformer contextual self-attention settings must be a LinformerSettings namedtuple'
+
         if type(n_local_attn_heads) is not tuple:
             n_local_attn_heads = tuple([n_local_attn_heads] * depth)
 
@@ -362,21 +315,31 @@ class LinearAttentionTransformer(nn.Module):
 
             parallel_net = Chunk(ff_chunks, FeedForward(dim), along_dim = 1) if not use_pkm else PKM(dim)
 
+            if linformer_settings is None:
+                attn = SelfAttention(dim, heads, causal, one_kv_head = one_kv_head, blindspot_size = blindspot_size, n_local_attn_heads = local_heads, local_attn_window_size = local_attn_window_size, psi_fn = psi_fn, dropout = attn_layer_dropout, attn_dropout= attn_dropout)
+            else:
+                attn = LinformerSelfAttention(dim, max_seq_len, heads = heads, one_kv_head = one_kv_head, dropout = attn_dropout, **linformer_settings._asdict())
+
             layers.append(nn.ModuleList([
-                PreNorm(dim, SelfAttention(dim, heads, causal, one_kv_head = one_kv_head, blindspot_size = blindspot_size, n_local_attn_heads = local_heads, local_attn_window_size = local_attn_window_size, psi_fn = psi_fn)),
+                PreNorm(dim, attn),
                 PreNorm(dim, parallel_net)
             ]))
 
             if attend_axially:
                 layers.append(nn.ModuleList([
-                    PreNorm(dim, FoldAxially(local_attn_window_size, SelfAttention(dim, heads, causal, one_kv_head = one_kv_head, psi_fn = psi_fn))),
-                    PreNorm(dim, Chunk(ff_chunks, FeedForward(dim), along_dim = 1))
+                    PreNorm(dim, FoldAxially(local_attn_window_size, SelfAttention(dim, heads, causal, one_kv_head = one_kv_head, psi_fn = psi_fn, dropout = attn_layer_dropout, attn_dropout= attn_dropout))),
+                    PreNorm(dim, Chunk(ff_chunks, FeedForward(dim, glu = ff_glu, dropout= ff_dropout), along_dim = 1))
                 ]))
 
             if receives_context:
+                if context_linformer_settings is None:
+                    attn = SelfAttention(dim, heads, one_kv_head = one_kv_head, psi_fn = psi_fn, dropout = attn_layer_dropout, attn_dropout= attn_dropout, receives_context = True)
+                else:
+                    attn = LinformerSelfAttention(dim, heads = heads, one_kv_head = one_kv_head, dropout = attn_dropout, **context_linformer_settings._asdict())
+
                 layers.append(nn.ModuleList([
-                    PreNorm(dim, SelfAttention(dim, heads, one_kv_head = one_kv_head, psi_fn = psi_fn, receives_context = True)),
-                    PreNorm(dim, Chunk(ff_chunks, FeedForward(dim), along_dim = 1))
+                    PreNorm(dim, attn),
+                    PreNorm(dim, Chunk(ff_chunks, FeedForward(dim, glu = ff_glu, dropout= ff_dropout), along_dim = 1))
                 ]))
 
         execute_type = ReversibleSequence if reversible else SequentialSequence
@@ -399,7 +362,7 @@ class LinearAttentionTransformer(nn.Module):
         return self.layers(x, **kwargs)
 
 class LinearAttentionTransformerLM(nn.Module):
-    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, causal = False, emb_dim = None, one_kv_head = False, reversible = False, ff_chunks = 1, blindspot_size = 1, n_local_attn_heads = 0, local_attn_window_size = 128, psi_fn = DEFAULT_PSI, return_embeddings = False, receives_context = False, pkm_layers = tuple(), pkm_num_keys = 128, attend_axially = False):
+    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, causal = False, emb_dim = None, one_kv_head = False, reversible = False, ff_chunks = 1, ff_glu = False, ff_dropout = 0., attn_layer_dropout = 0., attn_dropout = 0., blindspot_size = 1, n_local_attn_heads = 0, local_attn_window_size = 128, psi_fn = DEFAULT_PSI, return_embeddings = False, receives_context = False, pkm_layers = tuple(), pkm_num_keys = 128, attend_axially = False, linformer_settings = None, context_linformer_settings = None):
         assert (max_seq_len % local_attn_window_size) == 0, 'max sequence length must be divisible by the window size, to calculate number of kmeans cluster'
         super().__init__()
         emb_dim = default(emb_dim, dim)
@@ -407,7 +370,7 @@ class LinearAttentionTransformerLM(nn.Module):
 
         self.token_emb = nn.Embedding(num_tokens, emb_dim)
         self.axial_pos_emb = AxialPositionalEmbedding(emb_dim, axial_shape=(max_seq_len // local_attn_window_size, local_attn_window_size))
-        self.transformer = LinearAttentionTransformer(dim, depth, max_seq_len, heads = heads, causal = causal, one_kv_head = one_kv_head, ff_chunks = ff_chunks, reversible = reversible, blindspot_size = blindspot_size, n_local_attn_heads = n_local_attn_heads, local_attn_window_size = local_attn_window_size, psi_fn = psi_fn, receives_context = receives_context, pkm_layers = pkm_layers, pkm_num_keys = pkm_num_keys, attend_axially = attend_axially)
+        self.transformer = LinearAttentionTransformer(dim, depth, max_seq_len, heads = heads, causal = causal, one_kv_head = one_kv_head, ff_chunks = ff_chunks, ff_glu = ff_glu, ff_dropout = ff_dropout, attn_layer_dropout = attn_layer_dropout, attn_dropout = attn_dropout, reversible = reversible, blindspot_size = blindspot_size, n_local_attn_heads = n_local_attn_heads, local_attn_window_size = local_attn_window_size, psi_fn = psi_fn, receives_context = receives_context, pkm_layers = pkm_layers, pkm_num_keys = pkm_num_keys, attend_axially = attend_axially, linformer_settings = linformer_settings, context_linformer_settings = context_linformer_settings)
 
         if emb_dim != dim:
             self.transformer = ProjectInOut(self.transformer, emb_dim, dim, project_out = not return_embeddings)
